import glob
import os
import re
from itertools import product

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from cobaya.yaml import yaml_load_file
from getdist.paramnames import mergeRenames
from matplotlib.lines import Line2D

_default_root_path = "../output"


def _get_chain_filenames(path, prefix="mcmc", suffix=".txt"):
    return sorted(
        [
            f
            for f in glob.glob(os.path.join(path, f"{prefix}*{suffix}"))
            if re.match(rf"{prefix}(\.|)([0-9]+|){re.escape(suffix)}", os.path.basename(f))
        ]
    )


def _get_path(name, value):
    path = value
    if isinstance(value, dict):
        path = value.get("path", os.path.join(_default_root_path, str(name)))
    if not os.path.exists(path):
        raise ValueError(f"'{name} chains can not be located in '{path}'")
    return path


def create_symlink(mcmc_samples, prefix="mcmc"):
    """Create missing files when running chains without mpi support (as it is done in CC in2p3)"""
    regex = re.compile(rf".*{prefix}.*\.([0-9]+.txt)")
    for name, value in mcmc_samples.items():
        path = _get_path(name, value)

        if files := _get_chain_filenames(path, prefix=prefix):
            continue
        if not (files := _get_chain_filenames(path, prefix=prefix + ".*", suffix=".txt")):
            raise ValueError("Missing chain files!")

        print("Create MCMC symlinks...")
        for f in files:
            if m := regex.match(f):
                dest = f.replace(m.group(1), "")
                os.symlink(os.path.basename(f), dest + "txt")
                updated_yaml = os.path.join(os.path.dirname(f), f"{prefix}.updated.yaml")
                if not os.path.exists(updated_yaml):
                    os.symlink(os.path.basename(dest) + "updated.yaml", updated_yaml)


def print_chains_size(
    mcmc_samples,
    with_bar=False,
    bar_color="#b9b9b9",
    with_color_gradient=True,
    color_palette="Greens",
    hide_status=True,
    with_gelman_rubin=True,
    prefix="mcmc",
    mpi_run=True,
):
    """Print MCMC sample size given a set of directories

    Parameters
    ----------
    mcmc_samples: dict
      a dict holding a name as key for the sample and a corresponding directory as value
      or a dict configuration
    with_bar: bool
      showing an histogram bar for each cell given the number of mcmc samples
    bar_color: str
      bar color as hex string
    color_palette: str
      color palette to be used for background gradient
    hide_status: bool
      either to hide chain status (running, error, done)
    with_gelman_rubin: bool
      add Gelman-Rubin metric aka R-1
    prefix: str
      prefix for chain names (default is "mcmc.")
    mpi_run: bool
      mpi run flag (default true)
    """
    create_symlink(mcmc_samples, prefix)
    regex_log = re.compile(r".*mcmc\.([0-9]+).log")
    regex_progress = re.compile(r".*mcmc\.([0-9]+).progress")

    found_rminus1 = []
    data = {}
    status = dict(done="mcmc] The run has converged!", error="mcmc] *ERROR*")
    for irow, (name, value) in enumerate(mcmc_samples.items()):
        path = _get_path(name, value)
        name = value.get("label", name) if isinstance(value, dict) else name
        files = _get_chain_filenames(path, prefix=prefix, suffix=".log")
        if not files:
            print(f"Missing log files for chains '{name}' within path '{path}'!")
            return
        if len(files) == 1 and mpi_run:
            total_steps = {}
            data.setdefault(name, {})
            r = re.compile(
                r"\[(.*) : mcmc\] Progress @ (.*) : (.*) steps taken, and (.*) accepted."
            )
            with open(files[0]) as f:
                for line in f:
                    for state, msg in status.items():
                        if msg in line:
                            data[name].update({(mcmc_name, "status"): state})
                    found = r.findall(line)
                    if len(found) == 0:
                        continue
                    idx, time, current_steps, accepted_steps = found[0]
                    mcmc_name = f"mcmc {int(idx)+1}"
                    data[name].update({(mcmc_name, "status"): "running"})

                    current_steps = int(current_steps)
                    ts = total_steps.setdefault(idx, 0)
                    total_steps[idx] = current_steps if current_steps > ts else ts + current_steps
                    accepted_steps = int(accepted_steps)
                    rate = accepted_steps / total_steps[idx] if total_steps[idx] != 0 else None
                    data[name].update({(mcmc_name, "R-1"): None})
                    data[name].update({(mcmc_name, "accept."): accepted_steps})
                    data[name].update({(mcmc_name, "total"): total_steps[idx]})
                    data[name].update({(mcmc_name, "rate"): rate})
        else:
            r = re.compile(r"\[mcmc\] Progress @ (.*) : (.*) steps taken, and (.*) accepted.")
            for fn in sorted(files):
                total_steps = 0
                idx = 1 if not (m := regex_log.match(fn)) else m.group(1)
                mcmc_name = f"mcmc {idx}"
                data.setdefault(name, {}).update({(mcmc_name, "status"): "running"})
                with open(fn) as f:
                    for line in f:
                        for state, msg in status.items():
                            if msg in line:
                                data[name].update({(mcmc_name, "status"): state})
                        found = r.findall(line)
                        if len(found) == 0:
                            continue
                        time, current_steps, accepted_steps = found[0]

                        current_steps = int(current_steps)
                        total_steps = (
                            current_steps
                            if current_steps > total_steps
                            else total_steps + current_steps
                        )
                accepted_steps = int(accepted_steps)
                rate = accepted_steps / total_steps
                for field, content in zip(
                    ["R-1", "accept.", "total", "rate"], [None, accepted_steps, total_steps, rate]
                ):
                    data[name].update({(mcmc_name, field): content})
            if with_gelman_rubin:
                files = _get_chain_filenames(path, suffix=".progress")
                for fn in sorted(files):
                    idx = 1 if not (m := regex_progress.match(fn)) else m.group(1)
                    mcmc_name = f"mcmc {idx}"
                    with open(fn) as f:
                        for line in f:
                            pass
                        if line.startswith("#"):
                            continue
                        data[name].update({(mcmc_name, "R-1"): f"{float(line.split()[-2]):.2f}"})
                        found_rminus1 += [mcmc_name]

    df = pd.DataFrame.from_dict(data, orient="index")
    df.sort_index(level=0, axis=1, sort_remaining=False, inplace=True)
    df.dropna(axis=1, how="all", inplace=True)

    mcmc_names = list(df.columns.get_level_values(0).unique())

    # Append total count
    all_name = "all mcmc"
    dfs = [
        pd.DataFrame(
            {(all_name, field): df.loc[:, df.columns.get_level_values(1) == field].sum(axis=1)}
        )
        for field in ["accept.", "total"]
    ]
    df = pd.concat([df] + dfs, axis=1)
    df[(all_name, "rate")] = df[(all_name, "accept.")] / df[(all_name, "total")]
    all_mcmc_names = mcmc_names + [all_name]

    s = df.style
    if hide_status:
        s.hide([(name, "status") for name in mcmc_names], axis="columns")
    if with_bar:
        s.bar(
            subset=[(name, "accept.") for name in all_mcmc_names],
            color=bar_color,
            height=50,
            width=60,
        )
    if with_color_gradient:
        cm = sns.color_palette(color_palette, as_cmap=True)
        s.background_gradient(subset=[(name, "rate") for name in mcmc_names], cmap=cm, axis=None)
        # if with_gelman_rubin and found_rminus1:
        #     cm = sns.color_palette(color_palette + "_r", as_cmap=True)
        #     s.text_gradient(
        #         subset=[(name, "R-1") for name in mcmc_names if name in found_rminus1],
        #         cmap=cm,
        #         axis=None,
        #     )
        #     # s.highlight_null(color="white")
        s.background_gradient(
            subset=[(all_name, "total")], cmap=sns.color_palette("Blues", as_cmap=True)
        )
        s.background_gradient(
            subset=[(all_name, "rate")], cmap=sns.color_palette("Reds", as_cmap=True)
        )

    def _style_table(x):
        df = pd.DataFrame("", index=x.index, columns=x.columns)
        states = dict(running="#55A868", done="#4C72B0", error="#C44E52")
        for name, (state, color) in product(mcmc_names, states.items()):
            mask = x[(name, "status")] == state
            css = f"""color: {color}; text-decoration: {color} underline;
            text-decoration-thickness: 5px; font-weight: {'normal' if state=="running" else 'bold'}"""
            df.loc[mask, (name, "total")] = css
            if state == "error":
                df.loc[mask, (name, "rate")] = css

        return df

    s.format(
        {sub: "{:.1%}".format for sub in [(name, "rate") for name in all_mcmc_names]}, na_rep=""
    ).apply(_style_table, axis=None)
    return s


def print_results(samples, params, labels, limit=1):
    """Print results given a set of MCMC samples and a list of parameters

    Parameters
    ----------
    samples: list
      a list of MCSamples.
    params: list
      a list of parameters.
    labels: list
      a list of labels to be used as row index.
    limit: int
      the confidence limit of the results (default: 1 i.e. 68%).
    """
    params = params if isinstance(params, list) else list(params.keys())
    labels = labels if isinstance(labels, list) else list(labels.keys())
    d, cols = {}, {}

    merge_renames = {}
    for sample in samples:
        merge_renames = mergeRenames(
            sample.getParamNames().getRenames(keep_empty=True), merge_renames
        )
    reverse_renames = {vv: [k] for k, v in merge_renames.items() for vv in v}

    r = re.compile(r"(.*)(=|<|>)(.*)")
    for param in params:
        for sample in samples:
            latex = None
            sign = ""
            for par in param.split("|"):
                pars = [par] + merge_renames.get(par, []) + reverse_renames.get(par, [])
                for p in pars:
                    if sample.getParamNames().hasParam(p):
                        latex = sample.getInlineLatex(p, limit=limit)
                        if "<" in latex or ">" in latex:
                            latex = sample.getInlineLatex(p, limit=2)
                        break
                if latex:
                    break

            if not latex:
                value = " "
            else:
                # assert latex is not None, f"Parameter '{param}' not found!"
                found = r.findall(latex)
                assert (
                    len(found) == 1
                ), f"Something gets wrong when retrieving limits for '{param}' parameter!"
                name, sign, value = found[0]
                if param not in cols:
                    cols[param] = rf"${name.strip()}$"
            if "---" in value:
                value = "-"
            d.setdefault(param, []).append(f"${value}$" if sign == "=" else f"${sign}{value}$")
    df = pd.DataFrame(d, index=labels[: len(samples)])
    return df.rename(columns=cols)


def plot_chains(
    mcmc_samples,
    params,
    ncol=None,
    highlight_burnin=0.4,
    ignore_rows=0.0,
    show_mean_std=True,
    show_only_mcmc=None,
    no_cache=False,
    markers=None,
    markers_args=None,
    priors=None,
    priors_args=None,
    prefix="mcmc",
):
    """Plot MCMC sample evolution

        Parameters
        ----------
        mcmc_samples: dict
          a dict holding a name as key for the sample and a corresponding directory as value
          or a dict configuration
        params: dict or list
          a dict holding the parameter names for the different mcmc_samples or
          a unique list of parameter names
        ncol: int
          the number of columns within the plot
        highlight_burnin: float
          the fraction of samples to highlight (below the burnin value, the color is faded)
        ignore_rows: float
          the fraction of samples to ignore
        show_mean_std: bool
          show the mean/std values over the different samples
        show_only_mcmc: int or list
          only show chains given their number
        no_cache: bool
          remove the getdist cache
        markers: dict
    :      dictionnary holding the "expected" value for a parameter
        markers_args: dict
          marker kwargs to pass to plt.axhline
        prefix: str
          prefix name of chains
    """
    create_symlink(mcmc_samples, prefix)
    from getdist import loadMCSamples

    if not isinstance(params, (list, dict)):
        raise ValueError("Parameter list must be either a list or a dict!")

    if show_only_mcmc is not None:
        if not isinstance(show_only_mcmc, (int, list)):
            raise ValueError("Parameter 'show_only_mcmc' must be either a int or a list of int!")
        if isinstance(show_only_mcmc, int):
            show_only_mcmc = [show_only_mcmc]

    if ignore_rows > 0.0:
        highlight_burnin = 0.0

    markers = markers or {}
    markers_args = markers_args or dict(color="0.15", ls="--", lw=1)
    priors = priors or {}
    priors_args = priors_args or dict(color="0.75")
    stored_axes, color_palettes = {}, {}
    regex = re.compile(rf".*{prefix}\.([0-9]+).txt")
    merge_renames = {}
    for name, value in mcmc_samples.items():
        path = _get_path(name, value)
        if isinstance(value, dict):
            name = value.get("label", name)
        axes = None

        # Loop over files independently
        files = _get_chain_filenames(path, prefix=prefix)
        if not files:
            raise ValueError("Missing chain files!")

        chains = {}
        min_chain_size = np.inf
        sample = None
        for f in files:
            imcmc = 0 if not (m := regex.match(f)) else int(m.group(1))
            if show_only_mcmc and imcmc not in show_only_mcmc:
                continue
            kwargs = dict(no_cache=no_cache, settings={"ignore_rows": ignore_rows})
            try:
                sample = loadMCSamples(f[:-4], **kwargs)
                samples = sample.samples
            except AttributeError:
                sample = sample or loadMCSamples(os.path.join(os.path.dirname(f), prefix), **kwargs)
                samples = sample.getSeparateChains()[imcmc - 1].samples

            # Add renames
            merge_renames = mergeRenames(
                sample.getParamNames().getRenames(keep_empty=True), merge_renames
            )
            # Get param lookup table
            lookup = {}
            for i, par in enumerate(sample.getParamNames().names):
                lookup[par.name] = dict(pos=i, label=par.label)
                if renames := merge_renames.get(par.name):
                    lookup.update({r: lookup[par.name] for r in renames})

            if axes is None:
                # Keep only relevant parameters
                selected_params = [par for par in params if par in lookup]
                if ncol is None:
                    ncol, nrow = len(selected_params), 1
                else:
                    nrow = len(selected_params) // ncol
                    nrow += 1 if len(selected_params) > nrow * ncol else 0
                fig, axes = plt.subplots(nrow, ncol, sharex=True, figsize=(16, 2 * nrow))
                axes = axes.flatten()

            color = f"C{int(imcmc)-1}"
            # if color_name:
            #     if color_name not in color_palettes:
            #         color_palettes[color_name] = sns.blend_palette(
            #             ["white", color_name], n_colors=len(files) + 1
            #         )
            #     color = color_palettes[color_name][imcmc]
            if samples.shape[0] < min_chain_size:
                min_chain_size = samples.shape[0]
            for i, p in enumerate(selected_params):
                axes[i].set_ylabel(r"${}$".format(lookup[p].get("label")))
                y = samples[:, lookup[p].get("pos")]
                x = np.arange(len(y))
                idx_burnin = -int(highlight_burnin * len(y))
                axes[i].plot(x[idx_burnin:], y[idx_burnin:], alpha=0.75, color=color)
                if highlight_burnin > 0.0:
                    axes[i].plot(x[: idx_burnin + 1], y[: idx_burnin + 1], alpha=0.25, color=color)
                if p in markers:
                    axes[i].axhline(markers[p], **markers_args)
                if prior := priors.get(p):
                    mu, sigma = prior
                    axes[i].axhspan(mu - sigma, mu + sigma, **priors_args)
                chains.setdefault(p, []).append(y)

        if show_mean_std:
            for i, p in enumerate(selected_params):
                data = np.array([chain[:min_chain_size] for chain in chains[p]])
                mu, std = np.mean(data), np.std(data)
                axes[i].axhline(mu, color="0.6", lw=1)
                for sign in [-1, +1]:
                    axes[i].axhline(mu + std * sign, color="0.6", ls="--", lw=1)

        # Remove axes with no data inside
        _ = [fig.delaxes(ax) for ax in axes if not len(ax.get_lines())]
        leg = fig.legend(
            [Line2D([0], [0], color=f"C{int(f.split('.')[-2])-1}") for f in files],
            [f"mcmc #{f.split('.')[-2]}" for f in files],
            bbox_to_anchor=(0.5, 1.025),
            labelcolor="linecolor",
            loc="center",
            title=name,
            ncol=len(files),
            title_fontsize="large",
        )
        # leg._legend_box.align = "center"
        fig.tight_layout()
        stored_axes[name] = {p: axes[i] for i, p in enumerate(selected_params)}

    return stored_axes


def plot_progress(mcmc_samples, sharex=True, share_fig=False):
    """Plot Gelman R-1 parameter and acceptance rate

    Parameters
    ----------
    mcmc_samples: dict
      a dict holding a name as key for the sample and a corresponding directory as value.
    sharex: bool
      share the x-axis between the several plot progress (default: True)
    share_fig: bool
      plots are done within the same figure
    """
    nrows = len(mcmc_samples) if not share_fig else 1
    ncols = 2
    fig, axes = plt.subplots(nrows, ncols, figsize=(15, 3 * nrows), sharex=sharex)
    axes = np.atleast_2d(axes)

    regex = re.compile(r".*mcmc\.([0-9]+).progress")
    for i, (name, value) in enumerate(mcmc_samples.items()):
        path = _get_path(name, value)
        if isinstance(value, dict):
            name = value.get("label", name)

        files = _get_chain_filenames(path, suffix=".progress")
        for fn in files:
            with open(fn) as f:
                cols = [a.strip() for a in f.readline().lstrip("#").split()]
            df = pd.read_csv(
                fn, names=cols, comment="#", sep=" ", skipinitialspace=True, index_col=False
            )
            if df.N.empty:
                continue

            idx = 1 if not (m := regex.match(fn)) else m.group(1)
            color = f"C{int(idx)-1}" if not share_fig else f"C{i}"
            kwargs = dict(label=f"mcmc" + (f" #{idx}" if idx else ""), color=color, alpha=0.75)
            ix = i if not share_fig else 0
            axes[ix, 0].semilogy(df.N, df.Rminus1, "-o", **kwargs)
            axes[ix, 0].set_ylabel(r"$R-1$")

            axes[ix, 1].plot(df.N, df.acceptance_rate, "-o", **kwargs)
            axes[ix, 1].set_ylabel(r"acceptance rate")
        if len(files) > 1:
            leg = axes[i, 1].legend(
                title=name,
                bbox_to_anchor=(1, 1),
                loc="upper left",
                labelcolor="linecolor",
                alignment="left",
            )
    if share_fig:
        leg = fig.legend(
            [Line2D([0], [0], color=f"C{i}") for i, _ in enumerate(mcmc_samples)],
            mcmc_samples,
            bbox_to_anchor=(1, 1),
            labelcolor="linecolor",
            loc="center left",
        )
    fig.tight_layout()

    for i in range(nrows):
        if not axes[i, 0].lines:
            fig.delaxes(axes[i, 0])
            fig.delaxes(axes[i, 1])
    if not fig.axes:
        plt.close(fig)
    return fig


def get_sampled_parameters(mcmc_samples, prefix="mcmc", return_params=False):
    """Print MCMC sampled parameters

    Parameters
    ----------
    mcmc_samples: dict
      a dict holding a name as key for the sample and a corresponding directory as value
      or a dict configuration
    prefix: str
      prefix for chain names (default is "mcmc.")
    return_params: bool
      return dict of params for each MCMC chain (default false)
    """

    create_symlink(mcmc_samples, prefix)
    r1 = re.compile(r".*mcmc\] \*.*: (.*)")
    r2 = re.compile(r".*model\].*Input: (.*)")

    sampled_params = {}
    latex_table = {}
    for name, value in mcmc_samples.items():
        path = _get_path(name, value)
        name = value.get("label", name) if isinstance(value, dict) else name
        files = _get_chain_filenames(path, prefix=prefix, suffix=".log")
        if not files:
            print(f"Missing log files for chains '{name}' within path '{path}'!")
            return

        with open(files[0]) as f:
            for line in f:
                found = r1.findall(line) or r2.findall(line)
                if len(found) == 0:
                    continue
                params = found[0]
                sampled_params.setdefault(name, []).extend(eval(params))
                if "Sampling!" in line:
                    break

        updated_yaml = yaml_load_file(os.path.join(path, f"{prefix}.updated.yaml"))
        # To convert latex symbol to mathml
        # from latex2mathml.converter import convert
        # par: convert(rf"{meta.get('latex', par)}")
        latex_table |= {
            par: rf"${meta.get('latex', par)}$"
            for par, meta in updated_yaml.get("params", {}).items()
        }

    df = (
        pd.DataFrame.from_dict(sampled_params, orient="index")
        .T.fillna("")
        .map(lambda x: latex_table.get(x, ""))
    )
    dfs = df.style.set_properties(width="150px")
    if return_params:
        return dfs, sampled_params
    return dfs



import logging
import os

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from getdist import plots
from getdist.plots import get_subplot_plotter
from matplotlib.lines import Line2D
from scipy import stats

from .tools import _get_path


def set_style(
    use_seaborn=False,
    seaborn_style="ticks",
    seaborn_context="paper",
    palette="tab10",
    use_svg=False,
    use_tex=False,
    print_load_details=False,
    logging_level="error",
    backend="inline",
    **rc,
):
    """Set default plot settings

    Parameters
    ----------
    use_seaborn: bool
      use seaborn theming option
    seaborn_style: str
      set seaborn style (default: `ticks`)
    seaborn_context: str
      set seaborn context (default: `paper`)
    palette: str
      color palette name to be used for individual colors (default: tab10)
    use_svg: bool
      use `svg` output format for figure
    use_tex: bool
      use LaTeX axis labels
    print_load_details: bool
      print load details when getdist loads samples
    logging_level: str
      logging level to be passed to getdist
    backend: str
      matplotlib backend (default: inline)
    rc: dict
      overload matplotlib rc parameters
    """

    logging.getLogger("root").setLevel(getattr(logging, logging_level.upper(), logging.error))

    colors = None
    if palette == "the-lab":
        colors = dict(
            blue="#015692",
            orange="#b75501",
            green="#54790d",
            red="#c02d2e",
            purple="#803378",
            gray="#656e77",
        )
    elif palette == "getdist":
        colors = dict(blue="#006FED", red="#E03424", green="#009966")
    elif use_seaborn:
        palette = "deep"

    _default_color_cycle = [
        "blue",
        "orange",
        "green",
        "red",
        "purple",
        "brown",
        "pink",
        "gray",
        "yellow",
        "cyan",
    ]
    if colors:
        for color, code in colors.items():
            rgb = mpl.colors.colorConverter.to_rgb(color)
            mpl.colors.colorConverter.colors[code] = rgb
            mpl.colors.colorConverter.cache[code] = rgb

        _default_prop_cycle = mpl.rcParams["axes.prop_cycle"].by_key()["color"]
        mpl.rcParams["axes.prop_cycle"] = mpl.cycler(
            color=[
                colors.get(color, _default_prop_cycle[i])
                for i, k in enumerate(_default_color_cycle)
            ]
        )
    elif palette in ["tab10", "pastel", "muted", "bright", "deep", "colorblind", "dark"]:
        for code, color in zip(_default_color_cycle, sns.color_palette(palette)):
            rgb = mpl.colors.colorConverter.to_rgb(color)
            mpl.colors.colorConverter.colors[code] = rgb
            mpl.colors.colorConverter.cache[code] = rgb
    else:
        sns.set_palette(palette, n_colors=10)

    import getdist

    getdist.chains.print_load_details = print_load_details

    if use_tex and use_svg:
        print("Using latex text with svg output format does not play well. Disable the later one")
        use_svg = False

    if use_svg:
        import matplotlib_inline

        matplotlib_inline.backend_inline.set_matplotlib_formats("svg")

    rc = rc or {"axes.spines.top": False, "axes.spines.right": False}
    if use_tex:
        rc.update({"text.usetex": True})
    if use_seaborn:
        sns.set_theme(rc=rc, style=seaborn_style, context=None if use_tex else seaborn_context)
    else:
        plt.rcParams.update(rc)

    # Fix for jupyter hub @ NERSC
    get_ipython().run_line_magic("matplotlib", backend)


def get_default_settings(colors=None, linewidth=1, num_plot_contours=3):
    """Set default getdist plot settings

    Parameters
    ----------
    colors: list
      list of colors to be used
    linewidth: float
      line width of contours and KDE distributions (default: 2)
    num_plot_contours: int
      number of contour levels (default: 3)
    """
    if not colors:
        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]

    from getdist.plots import GetDistPlotSettings

    plot_settings = GetDistPlotSettings()
    plot_settings.solid_colors = colors
    plot_settings.line_styles = colors
    plot_settings.num_plot_contours = num_plot_contours
    plot_settings.linewidth = linewidth
    plot_settings.legend_fontsize = 15
    plot_settings.legend_colored_text = True
    plot_settings.scaling = False
    return plot_settings


def triangle_plot(*args, **kwargs):
    """Overloaded triangle_plot function with additional features"""
    g = get_subplot_plotter(settings=get_default_settings())
    g.triangle_plot(*args, **kwargs)

    if kwargs.get("despine", True):
        despine(g)

    if priors := kwargs.get("priors"):
        show_priors(
            g,
            priors,
            color=kwargs.get("prior_color", "gray"),
            with_legend=kwargs.get("prior_legend", True),
        )

    return g


def plots_1d(*args, **kwargs):
    """Overloaded plots_1d function with additional features"""
    default_plotter_options = {"width_inch": 4}
    plotter_kwargs = {k: kwargs.get(k, v) for k, v in default_plotter_options.items()}

    if legend_kwargs := kwargs.get("legend_kwargs"):
        legend_labels = kwargs.get("legend_labels")
        kwargs.update(dict(legend_labels=[]))

    g = get_subplot_plotter(settings=get_default_settings(), **plotter_kwargs)
    g.plots_1d(*args, **kwargs)

    if priors := kwargs.get("priors"):
        show_priors(
            g,
            priors,
            color=kwargs.get("prior_color", "gray"),
            with_legend=kwargs.get("prior_legend", True),
            legend_kwargs=dict(loc="best"),
        )

    if kwargs.get("despine", True):
        despine(g, all_axes=True)

    if legend_kwargs:
        g.add_legend(legend_labels, **legend_kwargs)

    return g


def plots_2d(*args, **kwargs):
    """Overloaded plots_2d function with additional features"""
    default_plotter_options = {"width_inch": 4}
    plotter_kwargs = {k: kwargs.get(k, v) for k, v in default_plotter_options.items()}

    g = get_subplot_plotter(settings=get_default_settings(), **plotter_kwargs)
    g.plots_2d(*args, **kwargs)

    if titles := kwargs.get("titles"):
        for title, ax in zip(titles, g.fig.axes):
            ax.set_title(title)

    return g


def get_mc_samples(
    mcmc_samples,
    prefix="mcmc",
    burnin=0.4,
    no_cache=False,
    as_dict=True,
    selected=None,
    excluded=None,
    select_first=None,
    no_progress_bar=True,
):
    """Print MCMC sample size given a set of directories

    Parameters
    ----------
    mcmc_samples: dict
      a dict holding a name as key for the sample and a corresponding directory as value
      or a dict configuration
    prefix: str
      filename prefix (default: 'mcmc')
    burnin: float
      burning fraction (default: 0.4)
    no_cache: bool
      either use or not the pickle cache file
    as_dict: bool
      return a dictionnary holding the kwargs to be used in getdist plot
    selected: list
      list of selected samples
    excluded: list
      list of excluded samples
    select_fist: str
      set the name of the first sample to return
    no_progress_bar: bool
      either enable or disable progress bar from tqdm
    """
    from getdist import loadMCSamples
    from tqdm.auto import tqdm

    selected = selected or list(mcmc_samples.keys())
    excluded = excluded or []

    if isinstance(selected, str):
        selected = [selected]
    if isinstance(excluded, str):
        excluded = [excluded]

    for e in excluded:
        selected.remove(e)

    if select_first:
        selected.remove(select_first)
        selected = [select_first] + selected

    default_prefix = prefix
    samples, labels, colors = [], [], []
    for name in (pbar := tqdm(selected, disable=no_progress_bar)):
        pbar.set_description(f"Loading '{name}'")
        value = mcmc_samples[name]
        if isinstance(value, str):
            value = dict(path=value)
        path = _get_path(name, value)
        prefix = value.get("prefix", default_prefix)
        labels += [value.get("label", name)]
        colors += [value.get("color", f"C{selected.index(name)}")]

        samples += [
            loadMCSamples(
                os.path.join(path, prefix),
                settings={"ignore_rows": burnin},
                no_cache=no_cache,
                chain_exclude=value.get("exclude"),
            )
        ]
    if as_dict:
        return samples, dict(legend_labels=labels, colors=colors, diag1d_kwargs={"colors": colors})
    return samples, labels, colors


def show_results(g, results, with_legend=True, legend_kwargs=None):
    """Show results values on a given set of axes
    Parameters
    ----------
    g: getdist.plots
      the getdist plotter instance
    results: dict
      dictionary holding result values. The dict can be either a parameter: loc/scale combo or
      parameter: dict(loc/scale +  and other settings such as color, linestyle)
    with_legend: bool
      add legend
    legend_kwargs: dict
      set legend kwargs
    """
    for name, result in results.items():
        show_priors(
            g,
            result.get("values", result),
            color=result.get("color"),
            ls=result.get("ls", "-"),
            with_legend=False,
        )
    kwargs = dict(
        obj=g.fig.axes[-1],
        labels=results.keys(),
        colors=[v.get("color") for v in results.values()],
        fontsize=10,
        bbox_to_anchor=(1, 1),
        loc="upper left",
    )
    kwargs |= legend_kwargs or {}
    if with_legend:
        add_legend(**kwargs)


def show_priors(g, priors, color="gray", ls="--", with_legend=True, legend_kwargs=None):
    """Show prior values on a given set of axes
    Parameters
    ----------
    g: getdist.plots
      the getdist plotter instance
    priors: dict
      dictionary holding loc/scale value for normal distribution
    color: str
      the color name
    ls: str
      the line style
    with_legend: bool
      add legend
    legend_kwargs: dict
      set legend kwargs
    """
    for par, val in priors.items():
        if not (ax := g.get_axes_for_params(par)):
            continue
        if isinstance(val, float):
            ax.axvline(val, color=color, ls=ls)
        else:
            x = np.linspace(*ax.get_xlim(), 100)
            y = stats.norm.pdf(x, *val)
            ax.plot(
                x, y / y.max(), color=color, ls=ls, label=rf"${ax.getdist_params[0].label}$ prior"
            )
        if with_legend:
            legend_kwargs = legend_kwargs or dict(loc="upper left", bbox_to_anchor=(1, 1))
            ax.legend(**legend_kwargs)


def show_tau_prior(g, loc=0.054, scale=0.0073, color="gray", ls="--"):
    """Dedicated function to plot tau prior

    Parameters
    ----------
    g: getdist.plots
      the getdist plotter instance
    loc: float
      the central tau value
    scale: float
      the scale/sigma of tau value
    color: str
      the color name
    ls: str
      the line style
    """

    show_priors(g, priors=dict(tau=(loc, scale)), color=color, ls=ls)


def despine(g, all_axes=False, **kwargs):
    """Special function to remove left spine in getdist plots

    Parameters
    ----------
    g: getdist.plots
      the getdist plotter instance
    all_axes: bool
      either to remove the spines for all the axes or not
    """
    kwargs = {"left": True} or kwargs
    if all_axes:
        sns.despine(fig=g.fig, **kwargs)
    else:
        sns.despine(ax=g.subplots[0, 0], **kwargs)


def plot_mean_values(samples, params, colors=None, palette=None, labels=None, figsize=None):
    """Plot mean and sigma values from posterior distributions

    Parameters
    ----------
    samples: list
      a list holding getdist MCSamples
    params: dict or list
      a dict holding the parameter names and its default value or
      a unique list of parameter names
    colors: list or str
      the colors of the different markers. If colors == "chi2" then the markers will be colored
      relatively to their chi2 values (if default values are given for parameters)
    palette: str
      the color palette to be used in case no colors are provided
    labels: list
      a list of labels describing each sample
    figsize: tuple
      the figure size
    """
    nparams, nsamples = len(params), len(samples)
    figsize = figsize or (24, 7)
    params = params if isinstance(params, dict) else {k: None for k in params}
    fig, axes = plt.subplots(nrows=1, ncols=nparams, sharey=True, figsize=figsize)
    fig.subplots_adjust(hspace=0, wspace=0.15)

    use_stats_color = None
    if colors is not None:
        if isinstance(colors, str):
            use_stats_color = colors
            colors = nsamples * [None if use_stats_color in ["chi2", "p-value"] else colors]
    else:
        colors = sns.color_palette(palette, n_colors=nsamples)
    cmap = sns.color_palette(palette or "flare", as_cmap=True)

    shape = (nparams, nsamples)
    chi2s, means, sigmas = np.full(shape, None), np.zeros(shape), np.zeros(shape)
    for i, sample in enumerate(samples):
        marge = sample.getMargeStats()
        latex = sample.getLatex(params)[0]
        for j, name in enumerate(params):
            par = marge.parWithName(name)
            x, xerr = par.mean, par.err
            means[j, i], sigmas[j, i] = x, xerr
            if params.get(name) is not None:
                chi2s[j, i] = ((x - params[name]) / xerr) ** 2
                if use_stats_color == "chi2":
                    colors[i] = cmap(chi2s[j, i])
                if use_stats_color == "p-value":
                    colors[i] = cmap(stats.chi2.sf(chi2s[j, i], 1))
            if axes[j].get_xlabel() == "":
                axes[j].set(xlabel=r"${}$".format(latex[j]), yticks=[])
            *_, bars = axes[j].errorbar(x, i, xerr=xerr, fmt="o", color=colors[i])
            for bar in bars:
                bar.set(alpha=0.5, color=colors[i], linewidth=3)

    # Customize axes and labels
    unique_color = "black" if len(np.unique(colors)) > 1 else colors[0]
    for j, name in enumerate(params):
        mu = np.average(means[j], weights=1 / sigmas[j] ** 2)
        sigma = np.mean(sigmas[j])
        axes[j].axvline(mu, color=unique_color, linestyle="--")
        axes[j].axvspan((mu - sigma), (mu + sigma), color="gray", alpha=0.15)
        if params.get(name) is not None:
            axes[j].spines["left"].set_position(("data", params[name]))
        else:
            axes[j].spines["left"].set_visible(False)
        if not np.all(chi2s[j] == None):
            pvalue = stats.chi2.sf(np.sum(chi2s[j]), nsamples)
            axes[j].set_title(f"$P(\chi^2)$ = {pvalue:.3f}", fontsize=10)

    # Add labels
    if labels is not None:
        ax = axes[0]
        for i, label in enumerate(labels):
            ax.text(
                ax.get_xlim()[0],
                i,
                label,
                va="center",
                ha="right",
                fontsize=8,
                color=unique_color if use_stats_color in ["chi2", "p-value"] else colors[i],
            )
    # Add X² colorbar
    if use_stats_color in ["chi2", "p-value"]:
        from matplotlib import cm, colors

        vmin = chi2s.min() if use_stats_color == "chi2" else 0.0
        vmax = chi2s.max() if use_stats_color == "chi2" else 1.0
        fig.colorbar(
            cm.ScalarMappable(norm=colors.Normalize(vmin=vmin, vmax=vmax), cmap=cmap),
            ax=axes,
            label=r"$\chi^2$" if use_stats_color == "chi2" else r"$p$-value",
            shrink=0.8,
        )


def plot_mean_distributions(samples, params, colors="0.7", return_results=False, nx=None):
    """Plot mean KDE distributions

    Parameters
    ----------
    samples: list
      a list holding getdist MCSamples
    params: dict or list
      a dict holding the parameter names and its default value or
      a unique list of parameter names
    colors: list or str
      the colors of the different markers. If colors == "chi2" then the markers will be colored
      relatively to their chi2 values (if default values are given for parameters)
    return_results: bool
      return a pandas Dataframe holding the results
    """
    from getdist.plots import get_subplot_plotter

    nsamples = len(samples)
    g = get_subplot_plotter(subplot_size=3, subplot_size_ratio=1.4)
    g.settings.line_styles = (
        nsamples * ["-" + colors] if isinstance(colors, str) else ["-" + color for color in colors]
    )
    nx = nx or nsamples
    g.plots_1d(samples, params, nx=nx, legend_labels=[], lws=2)

    results = {}
    axes = g.subplots.flatten()
    for i, name in enumerate(params):
        ax = axes[i]
        if not ax:
            continue
        means, sigmas = np.empty(nsamples), np.empty(nsamples)
        for j, sample in enumerate(samples):
            marge = sample.getMargeStats()
            par = marge.parWithName(name)
            means[j], sigmas[j] = par.mean, par.err
        mu = np.average(means, weights=1 / sigmas**2)
        sigma = np.mean(sigmas)
        x = np.linspace(*ax.get_xlim(), 1000)
        y = stats.norm.pdf(x, mu, sigma)
        ax.plot(x, y / y.max(), color="black", lw=2.5)
        if params.get(name) is not None:
            param = params.get(name)
            if isinstance(param, (tuple, list)):
                y = stats.norm.pdf(x, *param)
                ax.plot(x, y / y.max(), color="black", ls="--", lw=2.5)
            else:
                ax.axvline(param, color="red", lw=3)
            legend = ax.legend([], loc="upper right")
            legend.set_title(
                r"$\frac{{{:.1f}\,\sigma}}{{\sqrt{{N_{{\rm sim}}}}}}$".format(
                    (mu - param) / sigma * np.sqrt(nsamples)
                ),
                prop={"size": 16},
            )
        results[ax.get_xlabel()] = {"value rec.": mu, "$\sigma$ rec.": sigma}
        ax.spines.left.set_visible(False)

    if return_results:
        return pd.DataFrame.from_dict(results, orient="index")


def add_legend(obj, labels=None, colors=None, ls=None, **kwargs):
    """Add legend

    Parameters
    ----------
    obj: figure or axis
      a matplotlib figure or axis
    params: dict or list
      a dict holding the parameter names and its default value or
      a unique list of parameter names
    colors: list or str
      the colors of the different markers. If colors == "chi2" then the markers will be colored
      relatively to their chi2 values (if default values are given for parameters)
    return_results: bool
      return a pandas Dataframe holding the results
    """

    labels = labels or kwargs.get("legend_labels")

    colors = colors or [None for label in labels]
    ls = ls or ["-" for label in labels]
    handles = [Line2D([0], [0], color=colors[i], ls=ls[i]) for i, label in enumerate(labels)]

    leg = obj.legend(
        handles,
        labels,
        bbox_to_anchor=kwargs.get("bbox_to_anchor", (0.5, 1)),
        labelcolor=kwargs.get("labelcolor", "linecolor"),
        loc=kwargs.get("loc", "center"),
        title=kwargs.get("title"),
        ncol=kwargs.get("ncol", 1),
        alignment=kwargs.get("alignment", "left"),
        fontsize=kwargs.get("fontsize", 15),
    )


def plot_correlation_matrix(samples, params, nx=None, figsize=None):
    cmap = sns.diverging_palette(230, 20, as_cmap=True)
    samples = np.atleast_1d(samples)

    ncols = nx or len(samples)
    nrows = len(samples) // ncols
    figsize = figsize or 2 * [len(params) // 2]
    fig, axes = plt.subplots(ncols=ncols, nrows=nrows, figsize=figsize)
    for ax, sample in zip(np.atleast_2d(axes).flatten(), samples):
        latex = [rf"${tex}$" for tex in sample.getLatex(params)[0]]
        corr = sample.corr(params)
        matrix = pd.DataFrame(data=corr, index=latex, columns=latex)

        mask = np.triu(np.ones_like(matrix, dtype=bool))
        sns.heatmap(
            matrix,
            mask=mask,
            cmap=cmap,
            center=0,
            square=True,
            linewidths=0.5,
            cbar_kws={"shrink": 0.5},
            annot=True,
            annot_kws={"size": 6},
            fmt=".2f",
            vmin=-1,
            vmax=1,
            ax=ax,
            cbar=False,
        )



_cosmo_labels = dict(
    cosmomc_theta=r"\theta_\mathrm{MC}",
    logA=r"\log(10^{10} A_\mathrm{s})",
    As=r"A_\mathrm{s}",
    Alens=r"A_\mathrm{lens}",
    ns=r"n_\mathrm{s}",
    ombh2=r"\Omega_\mathrm{b}h^2",
    omch2=r"\Omega_\mathrm{c}h^2",
    tau=r"\tau_\mathrm{reio}",
)



import os
import tempfile
from copy import deepcopy

import numpy as np
import pandas as pd
import seaborn as sns

_default_packages_path = os.environ.get("COBAYA_PACKAGES_PATH") or os.path.join(
    tempfile.gettempdir(), "cobaya_utilities"
)

from getdist.types import NumberFormatter

_formatter = NumberFormatter()
_cached_fisher_matrix = None


def _get_sampled_params(params):
    sampled_params = deepcopy(params)
    sampled_params.update(
        {
            k: {"prior": {"min": 0.9 * v if v != 0 else -0.5, "max": 1.1 * v if v != 0 else +0.5}}
            for k, v in sampled_params.items()
        }
    )
    # Special treatment for logA and As
    if "logA" in params:
        sampled_params["logA"].update({"drop": True})
        sampled_params["As"] = {"value": "lambda logA: 1e-10*np.exp(logA)"}
    return sampled_params


def compute_fisher_matrix(
    likelihood_name,
    likelihood_config=None,
    params=None,
    return_correlation=False,
    epsilon=0.01,
    verbose=True,
    packages_path=None,
):
    """Compute and return the Fisher matrix

    Parameters
    ----------
    likelihood_name: str
      the name of the likelihood to be used
    likelihood_config: dict
      the likelihood config to give to cobaya
    params: dict
      the set of parameters to be used
    return_correlation: bool
      compute and return the correlation matrix rather than the covariance matrix
    epsilon: float
      the relative value to change the parameter when computing derivatives
    verbose: bool
      verbosity flag
    packages_path: str
      path to cobaya packages path
    """
    global _cached_fisher_matrix

    from cobaya.install import install
    from cobaya.log import get_logger
    from cobaya.model import get_model

    logger = get_logger("fisher")
    packages_path = packages_path or _default_packages_path

    likelihood_info = {"likelihood": {likelihood_name: likelihood_config}}
    install(likelihood_info, path=packages_path)

    params = params or {}
    info = {
        **likelihood_info,
        **{
            "params": _get_sampled_params(params),
            "theory": {"camb": {"extra_args": {"lens_potential_accuracy": 1}}},
        },
    }

    model = get_model(info, packages_path=packages_path)
    likelihood = model.likelihood[likelihood_name]
    theory = model.theory["camb"]

    # First grab the constant params and then update with the sampled one. We finally check for
    # missing parameters
    defaults = model.parameterization.constant_params()
    defaults.update({k: params.get(k) for k in model.parameterization.sampled_params().keys()})
    for k, v in defaults.items():
        if v is None:
            raise ValueError(f"Parameter '{k}' must be set!")

    deriv = {}
    for param in params:

        def _get_power_spectra(epsilon):
            point = defaults.copy()
            point.update(
                {
                    param: (
                        point[param] * (1 + epsilon)
                        if point[param] != 0
                        else point[param] + epsilon
                    )
                }
            )
            model.logposterior(point)
            return likelihood._get_power_spectra(theory.get_Cl(ell_factor=True), **point)

        delta = (_get_power_spectra(+epsilon) - _get_power_spectra(-epsilon)) / 2 / epsilon
        if defaults[param] != 0:
            delta /= defaults[param]

        if np.all(delta == 0):
            logger.warning(
                f"Sampling a parameter '{param}' that do not have "
                "any effect on power spectra! You should remove it from "
                "cobaya parameter dictionary."
            )
            continue

        deriv[param] = delta
        if verbose:
            logger.info(f"Computing parameter '{param}' done")

    fisher_params = list(deriv.keys())
    nparams = len(fisher_params)
    fisher_matrix = np.empty((nparams, nparams))
    for i1, p1 in enumerate(fisher_params):
        for i2, p2 in enumerate(fisher_params):
            fisher_matrix[i1, i2] = deriv[p1] @ likelihood.inv_cov @ deriv[p2]
    _cached_fisher_matrix = np.linalg.inv(fisher_matrix)

    from cobaya_utilities.utilities import _cosmo_labels

    labels = [
        f"${_cosmo_labels.get(name, model.parameterization.labels().get(name))}$"
        for name in fisher_params
    ]
    values = np.array(list(params.values()))
    sigmas = np.sqrt(np.diag(_cached_fisher_matrix))
    signal_over_noise = values / sigmas

    format_array = lambda array: [_formatter.formatNumber(n) for n in array]
    # [format_array(values), format_array(sigmas), format_array(signal_over_noise)]
    summary = pd.DataFrame(
        data=np.array([values, sigmas, signal_over_noise]).T,
        index=labels,
        columns=["value", r"$\sigma$", "S/N"],
    )
    summary["param"] = list(params.keys())
    _cached_fisher_matrix = pd.DataFrame(data=_cached_fisher_matrix, index=labels, columns=labels)
    if verbose:
        logger.info(f"Computing fisher matrix done")

    if return_correlation:
        return summary, _cached_fisher_matrix.div(np.outer(sigmas, sigmas))
    return summary


def plot_fisher_matrix(matrix=None, use_relplot=True, reset_cache=False, **matrix_args):
    """Compute and return the Fisher matrix

    Parameters
    ----------
    matrix: pandas.Dataframe
      the Fisher matrix to be plotted. In case no matrix is provided then
      the function will call the `compute_fisher_matrix` function
    use_relplot: bool
      use a scatter plot representation rather than a matrix view
    matrix_args: dict
      the paramaters to pass to the `compute_fisher_matrix` function
    reset_cache: bool
      recompute previous Fisher estimation
    """
    global _cached_fisher_matrix
    if reset_cache:
        _cached_fisher_matrix = None

    if _cached_fisher_matrix is not None:
        sigmas = np.sqrt(np.diag(_cached_fisher_matrix))
        matrix = _cached_fisher_matrix / np.outer(sigmas, sigmas)

    if matrix is None:
        matrix_args = matrix_arags.update({"return_correlation": True})
        summary, matrix = compute_fisher_matrix(**matrix_args)

    cmap = sns.diverging_palette(230, 20, as_cmap=True)
    if use_relplot:
        mask = np.triu(np.ones_like(matrix, dtype=bool), k=1)
        corr_mat = matrix.mask(~mask).stack().reset_index(name="correlation")
        with sns.axes_style("whitegrid"):
            g = (
                sns.relplot(
                    data=corr_mat,
                    x="level_0",
                    y="level_1",
                    hue="correlation",
                    size="correlation",
                    palette=cmap,
                    hue_norm=(-1, 1),
                    edgecolor="0.7",
                    height=8,
                    sizes=(50, 200),
                    size_norm=(-1, 1),
                )
                .set(xlabel="", ylabel="", aspect="equal")
                .set_xticklabels(rotation=90)
                .despine(left=True, bottom=True)
            )
            for artist in g.legend.legend_handles:
                artist.set(markeredgecolor="0.7", markeredgewidth=1)
    else:
        mask = np.triu(np.ones_like(matrix, dtype=bool))
        sns.heatmap(
            matrix,
            mask=mask,
            cmap=cmap,
            center=0,
            square=True,
            linewidths=0.5,
            cbar_kws={"shrink": 0.5},
        )
    return g.fig.axes


def generate_yaml_config(
    summary=None,
    sigma_scale=10,
    proposal_scale=0.1,
    with_ref=False,
    ref_scale=1,
    only_params=None,
    filename=None,
    error_if_exists=True,
    **matrix_args,
):
    """Generate cobaya yaml configuration for the sampled parameters

    Parameters
    ----------
    summary: pandas.Dataframe
      the summary product from `compute_fisher_matrix`. In case no summary is provided then
      the function will call the `compute_fisher_matrix` function
    sigma_scale: float
      the number of sigma for min/max values of the prior
    proposal_scale: float
      the proposal is equal to `proposal_scale * sigma`
    with_ref: bool
      generate a reference value normally distributed around the central value
      and sigma equal to `sigma * ref_scale` (see below for ref_scale)
    ref_scale: float
      the normal distribution has scale equal to `ref_scale * sigma`
    only_params: list
      generate yaml configuration only for this set of parameters
    filename: str
      the filename to dump the yaml configuration
    error_if_exists: bool
      raise an error if the yaml filename already exists otherwise erase its content
    matrix_args: dict
      the paramaters to pass to the `compute_fisher_matrix` function
    """

    if summary is None:
        summary = compute_fisher_matrix(**matrix_args)

    fn = lambda n: float(_formatter.formatNumber(n))

    yaml_dict = {}
    for latex, fields in summary.to_dict(orient="index").items():
        name = fields.get("param")
        if only_params and name not in only_params:
            continue
        value = float(fields.get("value"))
        sigma = float(fields.get(r"$\sigma$"))
        latex = latex.replace("$", "")
        prior_min = value - sigma_scale * sigma
        prior_max = value + sigma_scale * sigma
        yaml_dict.update({name: {"prior": {"min": fn(prior_min), "max": fn(prior_max)}}})
        if with_ref:
            yaml_dict[name].update(
                {"ref": {"dist": "norm", "loc": value, "scale": fn(sigma * ref_scale)}}
            )
        yaml_dict[name].update({"proposal": fn(sigma * proposal_scale), "latex": latex})

    from cobaya.yaml import yaml_dump, yaml_dump_file

    if filename:
        yaml_dump_file(filename, yaml_dict, error_if_exists=error_if_exists)
    else:
        print(yaml_dump(yaml_dict))



from setuptools import find_packages, setup

import versioneer

with open("README.rst") as readme_file:
    readme = readme_file.read()

setup(
    name="cobaya_utilities",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(),
    description="A set of functions to deal with MCMC output from cobaya",
    long_description=readme,
    long_description_content_type="text/x-rst",
    url="https://github.com/xgarrido/cobaya_utilities.git",
    author="Xavier Garrido",
    author_email="xavier.garrido@lal.in2p3.fr",
    keywords=["cobaya", "MCMC", "plot"],
    python_requires=">=3.9",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Education",
    ],
    install_requires=["cobaya", "seaborn", "tqdm"],
)
